from scipy.special import binom as bn
import numpy as np
from scipy.linalg import lstsq
from scipy.interpolate import interp1d

class bernstein(object):
    
    def __init__(self,t,n=5):
        self.t = t
        self.n = n
        self.knots()
        
    def basis(self,v):
        return bn(self.n,v)*self.t**v*(1-self.t)**(self.n-v)
    
    def knots(self,v=[]):
        self.A = np.array(np.zeros((len(self.t),self.n+1)))
        if len(v) == 0: v = range(self.n+1)
        v = np.array(v)
        v[v>self.n] = self.n
        v[v<0] = 0
        for i,vv in enumerate(v):
            self.A[:,i] = self.basis(vv)
            
    def fit(self,data):
        self.data = data
        b = lstsq(self.A,data)[0]
        return self.gen(b),b
        
    def spline(self,b):
        return interp1d(self.t,self.gen(b),fill_value=0,bounds_error=False)
        
    def gen(self,b):
        if len(b) == 2*(self.n+1):
            self.knots(v=b[self.n+1:])
        return np.dot(self.A,b[:self.n+1])
        
    def error(self,b):
        data = self.data
        poly = self.gen(b)
        rms = np.sqrt(np.mean((data-poly)**2))
        return rms


import numpy as np
from scipy.interpolate import interp1d
from scipy.interpolate import UnivariateSpline as spline
import pylab as pl
import scipy as sp
from matplotlib.collections import PolyCollection
import collections
from warnings import warn

def elipsoid(theta,ro,zo,A,k,delta):
    R = ro+ro/A*np.cos(theta+delta*np.sin(theta))
    Z = zo+ro/A*k*np.sin(theta)
    return R,Z

def loop_vol(R,Z,plot=False):
    imin,imax = np.argmin(Z),np.argmax(Z)
    Rin = np.append(R[::-1][:imin+1][::-1],R[:imin+1])
    Zin = np.append(Z[::-1][:imin+1][::-1],Z[:imin+1])
    Rout = R[imin:imax+1]
    Zout = Z[imin:imax+1]
    if plot:
        pl.plot(R[0],Z[0],'bo')
        pl.plot(R[20],Z[20],'bd')
        pl.plot(Rin,Zin,'bx-')
        pl.plot(Rout,Zout,'gx-')
    return vol_calc(Rout,Zout)-vol_calc(Rin,Zin)
    
def vol_calc(R,Z):
    dR = np.diff(R)
    dZ = np.diff(Z)
    V = 0
    for r,dr,dz in zip(R[:-1],dR,dZ):
        V += np.abs((r+dr/2)**2*dz)
    V *= np.pi
    return V
    
def unique(R,Z):
    L = length(R,Z)
    io = np.append(np.diff(L)>0,True)  # remove duplicates
    return R[io],Z[io],L[io]
    
def order(R,Z,anti=True):
    rc,zc = (np.mean(R),np.mean(Z))  
    theta = np.unwrap(np.arctan2(Z-zc, R-rc))
    if theta[-1]<theta[0]:
        R,Z = R[::-1],Z[::-1]
    if not anti:
        R,Z = R[::-1],Z[::-1]
    return R,Z
    
def clock(R,Z,reverse=True):  # order loop points in anti-clockwise direction
    rc,zc = (np.mean(R),np.mean(Z))  
    radius = ((R-rc)**2+(Z-zc)**2)**0.5
    theta = np.arctan2(Z-zc, R-rc)
    index = theta.argsort()[::-1]
    radius,theta = radius[index],theta[index] 
    R,Z = rc+radius*np.cos(theta),zc+radius*np.sin(theta)
    R,Z = np.append(R,R[0]),np.append(Z,Z[0])
    R,Z = rzSLine(R,Z,npoints=len(R)-1)
    if reverse:
        R,Z = R[::-1],Z[::-1]
    return R,Z
    
def theta_sort(R,Z,origin='lfs',**kwargs):
    xo = kwargs.get('xo',(np.mean(R),np.mean(Z)))
    anti = kwargs.get('anti',False)
    if origin == 'lfs':
        theta = np.arctan2(Z-xo[1],R-xo[0])
    elif origin == 'top':
        theta = np.arctan2(xo[0]-R,Z-xo[1]) 
    elif origin == 'bottom':
        theta = np.arctan2(R-xo[0],xo[1]-Z) 
    if kwargs.get('unwrap',False):
        theta = np.unwrap(theta)
    index = np.argsort(theta)
    R,Z = R[index],Z[index]
    if anti:
        R,Z = R[::-1],Z[::-1]
    return R,Z

def rt(R,Z,xo):
    theta = np.unwrap(np.arctan2(Z-xo[1],R-xo[0]))
    radius = np.sqrt((Z-xo[1])**2+(R-xo[0])**2)
    index = np.argsort(theta)
    radius,theta = radius[index],theta[index]
    return radius,theta
    
def rz(radius,theta,xo):
    R = xo[0]+radius*np.cos(theta)
    Z = xo[1]+radius*np.sin(theta)
    return R,Z

def rzSpline(R,Z,xo,npoints=500,w=None,s=0.005):
    radius,theta = rt(R,Z,xo)
    Ts = np.linspace(theta[0],theta[-1],npoints)
    if w is None:
        radius = spline(theta,radius,s=s)(Ts)
    else:
        radius = spline(theta,radius,w=w,s=s)(Ts)
    Rs,Zs = rz(radius,Ts,xo)
    return Rs,Zs,Ts
        
def rzSLine(R,Z,npoints=500,s=0,Hres=False):
    L = length(R,Z)
    if Hres: npoints *= 10
    Linterp = np.linspace(0,1,npoints)
    if s == 0:
        R = interp1d(L,R)(Linterp)
        Z = interp1d(L,Z)(Linterp)
    else:
        R = spline(L,R,s=s)(Linterp)
        Z = spline(L,Z,s=s)(Linterp)
    return R,Z
        
def rzInterp(R,Z,npoints=500,ends=True):
    L = length(R,Z)
    Linterp = np.linspace(0,1,npoints,endpoint=ends)
    R = interp1d(L,R)(Linterp)
    Z = interp1d(L,Z)(Linterp)
    return R,Z
    
def rzfun(R,Z):  # return interpolation functions
    L = length(R,Z)
    R = interp1d(L,R)
    Z = interp1d(L,Z)
    return R,Z
        
def rzCirc(R,Z):
    radius,theta = rt(R,Z)
    R,Z = rz(radius,theta)
    return R,Z
        
def length(R,Z,norm=True):
    L = np.append(0,np.cumsum(np.sqrt(np.diff(R)**2+np.diff(Z)**2)))
    if norm: L = L/L[-1]
    return L
    
def vector_length(X,norm=True):
    L = np.append(0,np.cumsum(np.sqrt(np.diff(X[:,0])**2+
                                      np.diff(X[:,1])**2+
                                      np.diff(X[:,2])**2)))
    if norm: L = L/L[-1]
    return L
    
def space(R,Z,npoints):
    L = length(R,Z)
    l = np.linspace(0,1,npoints)
    R,Z = interp1d(L,R)(l),interp1d(L,Z)(l)
    return R,Z
    
def rotate(theta,axis='z'):
    if axis == 'z':
        R = np.array([[np.cos(theta),-np.sin(theta),0],
                       [np.sin(theta),np.cos(theta),0],
                       [0,0,1]])
    elif axis == 'y':
        R = np.array([[np.cos(theta),0,-np.sin(theta)],
                       [0,1,0],
                       [np.sin(theta),0,np.cos(theta)]])
    elif axis == 'x':
        R = np.array([[1,0,0],
                      [0,np.cos(theta),-np.sin(theta)],
                      [0,np.sin(theta),np.cos(theta)]])        
    else:
        errtxt = 'incorrect roation axis {}'.format(axis)
        errtxt += ', select from [\'x\',\'y\',\'z\']'
        raise ValueError(errtxt)
    return R
    
def rotate2D(theta,xo=0,yo=0,dx=0,dy=0):
    R = np.array([[np.cos(theta),-np.sin(theta),0],
                  [np.sin(theta),np.cos(theta),0],
                  [0,0,1]])
    To = np.array([[1,0,-xo],[0,1,-yo],[0,0,1]])
    T1 = np.array([[1,0,xo+dx],[0,1,yo+dy],[0,0,1]])
    TRT = np.dot(T1,np.dot(R,To))[:2,:]
    return R[:2,:2],TRT
    
def rotate_vector2D(V,theta):
    Vmag = np.sqrt(V[0]**2+V[1]**2)
    R = rotate2D(theta)[0]
    Vhat = (R*np.matrix([[V[0]],[V[1]]])).getA1()/Vmag
    return Vhat
    
def normal(R,Z):
    dR,dZ = np.gradient(R),np.gradient(Z)
    mag = np.sqrt(dR**2+dZ**2)
    index = mag>0
    dR,dZ,mag = dR[index],dZ[index],mag[index]  # clear duplicates
    R,Z = R[index],Z[index]
    t = np.zeros((len(dR),3))
    t[:,0],t[:,1] = dR/mag,dZ/mag
    n = np.cross(t, [0,0,1])
    nR,nZ = n[:,0],n[:,1]
    return nR,nZ,R,Z
    
def inloop(Rloop,Zloop,R,Z,side='in'):
    if side == 'in':
        sign = 1
    elif side == 'out':
        sign = -1
    else:
        raise ValueError('define side, \'in\' or \'out\'')
    Rloop,Zloop = clock(Rloop,Zloop)
    nRloop,nZloop,Rloop,Zloop = normal(Rloop,Zloop)
    Rin,Zin = np.array([]),np.array([])
    if isinstance(R,collections.Iterable):
        for r,z in zip(R,Z):
            i = np.argmin((r-Rloop)**2+(z-Zloop)**2)
            dr = [Rloop[i]-r,Zloop[i]-z]  
            dn = [nRloop[i],nZloop[i]]
            if sign*np.dot(dr,dn) > 0:
                Rin,Zin = np.append(Rin,r),np.append(Zin,z)
        return Rin,Zin
    else:
        i = np.argmin((R-Rloop)**2+(Z-Zloop)**2)
        dr = [Rloop[i]-R,Zloop[i]-Z]  
        dn = [nRloop[i],nZloop[i]]
        return sign*np.dot(dr,dn) > 0
    
def max_steps(dR,dr_max):
    dRbar = np.mean(dR)
    nr=int(np.ceil(dRbar/dr_max))
    if nr < 2: nr = 2
    dr=dR/nr
    return dr,nr
    
def offset(R,Z,dR,min_steps=5,close_loop=False):
    R,Z = order(R,Z)  # enforce anti-clockwise
    dr_max = np.mean(dR)/min_steps  # maximum step size
    if np.mean(dR) != 0:
        dr,nr = max_steps(dR,dr_max)
        for i in range(nr):
            nR,nZ,R,Z = normal(R,Z)
            R = R+dr*nR    
            Z = Z+dr*nZ 
            if close_loop:
                R[0],Z[0] = np.mean([R[0],R[-1]]),np.mean([Z[0],Z[-1]])
                R[-1],Z[-1] = R[0],Z[0]
    return R,Z 

def cut(r,z):  # trim gaps
    for _ in range(len(r)-1):
        dl = np.diff(length(r,z))
        if np.max(dl) > 2*np.median(dl):
            icut = np.argmax(dl)-1
            if icut > len(dl)/2:
                r,z = r[:icut],z[:icut]
            else:
                r,z = r[icut:],z[icut:]
        else:
            break
    return r,z
   
class Loop(object):
    
    def __init__(self,R,Z,**kwargs):
        self.R = R
        self.Z = Z
        self.xo = kwargs.get('xo',(np.mean(R),np.mean(Z)))
    
    def rzPut(self):
        self.Rstore,self.Zstore = self.R,self.Z
        
    def rzGet(self):
        self.R,self.Z = self.Rstore,self.Zstore
        
    def fill(self,trim=None,dR=0,dt=0,ref_o=4/8*np.pi,dref=np.pi/4,
             edge=True,ends=True,
             color='k',label=None,alpha=0.8,referance='theta',part_fill=True,
             loop=False,s=0,gap=0,plot=False):
        dt_max = 0.1  # 2.5
        if not part_fill:
            dt_max = dt
        if isinstance(dt,list):
            dt = self.blend(dt,ref_o=ref_o,dref=dref,referance=referance,
                            gap=gap)
        dt,nt = max_steps(dt,dt_max)
        Rin,Zin = offset(self.R,self.Z,dR)  # gap offset
        for i in range(nt):
            self.part_fill(trim=trim,dt=dt,ref_o=ref_o,dref=dref,
                           edge=edge,ends=ends,color=color,label=label,alpha=alpha,
                           referance=referance,loop=loop,s=s,plot=False)
        Rout,Zout = self.R,self.Z
        if plot:
            polyparrot({'r':Rin,'z':Zin},{'r':Rout,'z':Zout},
                       color=color,alpha=1)  # fill
        return Rout,Zout
             
    def part_fill(self,trim=None,dt=0,ref_o=4/8*np.pi,dref=np.pi/4,
             edge=True,ends=True,
             color='k',label=None,alpha=0.8,referance='theta',loop=False,
             s=0,plot=False):
        Rin,Zin = self.R,self.Z
        if loop:
            Napp = 5  # Nappend
            R = np.append(self.R,self.R[:Napp])
            R = np.append(self.R[-Napp:],R)
            Z = np.append(self.Z,self.Z[:Napp])
            Z = np.append(self.Z[-Napp:],Z)
            R,Z = rzSLine(R,Z,npoints=len(R),s=s)
            if isinstance(dt,(np.ndarray,list)):
                dt = np.append(dt,dt[:Napp])
                dt = np.append(dt[-Napp:],dt)
            Rout,Zout = offset(R,Z,dt)
            print('part fill')
            Rout,Zout = Rout[Napp:-Napp],Zout[Napp:-Napp]
            Rout[-1],Zout[-1] = Rout[0],Zout[0]
        else:
            R,Z = rzSLine(self.R,self.Z,npoints=len(self.R),s=s)
            Rout,Zout = offset(R,Z,dt)
        self.R,self.Z = Rout,Zout  # update
        if trim is None:
            Lindex = [0,len(Rin)]
        else:
            Lindex = self.trim(trim)
        if plot:
            flag = 0
            for i in np.arange(Lindex[0],Lindex[1]-1):
                Rfill = np.array([Rin[i],Rout[i],Rout[i+1],Rin[i+1]])
                Zfill = np.array([Zin[i],Zout[i],Zout[i+1],Zin[i+1]])
                if flag is 0 and label is not None:
                    flag = 1
                    pl.fill(Rfill,Zfill,facecolor=color,alpha=alpha,
                            edgecolor='none',label=label)
                else:
                    pl.fill(Rfill,Zfill,facecolor=color,alpha=alpha,
                            edgecolor='none')
             
    def blend(self,dt,ref_o=4/8*np.pi,dref=np.pi/4,gap=0,referance='theta'):
        if referance is 'theta':
            theta = np.arctan2(self.Z-self.xo[1],self.R-self.xo[0])-gap
            theta[theta>np.pi] = theta[theta>np.pi]-2*np.pi
            tblend = dt[0]*np.ones(len(theta))  # inner 
            tblend[(theta>-ref_o) & (theta<ref_o)] = dt[1]  # outer 
            if dref > 0:
                for updown in [-1,1]:
                    blend_index = (updown*theta>=ref_o) &\
                                    (updown*theta<ref_o+dref)
                    tblend[blend_index] = dt[1]+(dt[0]-dt[1])/dref*\
                                        (updown*theta[blend_index]-ref_o)
        else:
            L = length(self.R,self.Z)
            tblend = dt[0]*np.ones(len(L))  # start
            tblend[L>ref_o] = dt[1]  # end
            if dref > 0:
                blend_index = (L>=ref_o) & (L<ref_o+dref)
                tblend[blend_index] = dt[0]+(dt[1]-dt[0])/dref*(L[blend_index]-
                                                                ref_o)
        return tblend
        
    def trim(self,trim,R,Z):
        L = length(R,Z,norm=True)
        index = []
        for t in trim:
            index.append(np.argmin(np.abs(L-t)))
        return index
    
def split_loop(r,z,xo,half):
    if 'upper' in half:
        index = z >= xo[1]
    elif 'lower' in half:
        index = z <= xo[1]
    else:
        errtxt = '\n'
        errtxt += 'specify loop segment [\'upper\',\'lower\']\n'
        raise ValueError(errtxt)
    r,z = r[index],z[index]
    r,z = theta_sort(r,z,xo=xo)  
    return r,z
   
def trim_loop(r,z):
    n = len(r)
    index = np.ones(n,dtype=bool)
    for i in range(n-2):
        if r[i+1] <= r[i]:
            index[i+1] = False
        else:
            index[i+1] = True  
        if index[i+1] and not index[i]:  # keep corner
            index[i] = True
    r,z = r[index],z[index]
    if r[0] > r[1]:
        r[0] = r[1]-1e-6
    if r[-1] < r[-2]:
        r[-1] = r[-2]+1e-6
    return r,z
    
def process_loop(r,z):
    xo = (np.mean(r),z[np.argmax(r)])
    r1,z1 = split_loop(r,z,xo,'upper')
    ro,zo = split_loop(r,z,xo,'lower')
    r1 = np.append(ro[-1],r1)  # join upper
    r1 = np.append(r1,ro[0])
    z1 = np.append(zo[-1],z1)
    z1 = np.append(z1,zo[0])
    ro,zo = trim_loop(ro,zo)
    r1,z1 = trim_loop(r1[::-1],z1[::-1])
    return (ro,zo),(r1,z1)
    
def read_loop(part,loop,npoints=100,close=True):
    r,z = part[loop]['r'],part[loop]['z']
    if len(r) > 0:
        r,z = theta_sort(r,z)  # sort azimuth
        if close:
            r,z = np.append(r,r[0]),np.append(z,z[0])  # close loop
    return r,z
    
def polyloop(xin,xout,color=0.5*np.ones(3),alpha=1):  # pair to single 
    x = {}
    for var in ['r','z']:
        x[var] = np.append(xin[var],xout[var][::-1])
        x[var] = np.append(x[var],xin[var][0])
    return x['r'],x['z']
      
def polyfill(r,z,color=0.5*np.ones(3),alpha=1):
    verts = np.array([r,z])
    verts = [np.swapaxes(verts,0,1)]
    coll = PolyCollection(verts,edgecolors='none',color=color,alpha=alpha)
    ax = pl.gca()
    ax.add_collection(coll)
    ax.autoscale_view()
    
def polyparrot(xin,xout,color=0.5*np.ones(3),alpha=1):  # polyloopfill
    r,z = polyloop(xin,xout)
    polyfill(r,z,color=color,alpha=alpha)
        
def pointloop(r,z,ref='max'):
    n = len(r)
    r_,z_ = np.zeros(n),np.zeros(n)
    if ref=='max':
        i = np.argmax(sp.linalg.norm([r,z],axis=0))
    else:
        i = np.argmin(z)
    r_[0],z_[0] = r[i],z[i]
    r,z = np.delete(r,i),np.delete(z,i)
    for i in range(n-1):
        dr = sp.linalg.norm([r-r_[i],z-z_[i]],axis=0)
        j = np.argmin(dr)
        r_[i+1],z_[i+1] = r[j],z[j]
        r,z = np.delete(r,j),np.delete(z,j)
    r,z = r_,z_
    r,z = np.append(r,r[0]),np.append(z,z[0])
    return r,z
        



import numpy as np

def OOB(var, delta=1e-3):  # out of bounds
    var = np.append(var[0]-delta, var)
    var = np.append(var, var[-1]+delta)
    return (var,len(var))
    
def edge(index, Narray):
    on_edge = False    
    for i,N in zip(index,Narray):
        if i==0 or i==N-1:
            on_edge = True
    return on_edge

class table(object):
    def __init__(self,filename,ext='.lib'):
        f = open(filename+ext, 'w')         
        self.f = f
        
    def nop(self):
        self.f.write('/nop\n')  # supress output
        
    def go(self):
        self.f.write('/go\n')  # resume output
        
    def load(self, name, data, index=[]):
        if len(data) == 1:
            data = np.array([data[0],data[0]])
            index = [np.array([index[0][0]-1e-34,index[0][0]+1e-34])]
        self.name = name
        self.data = data
        self.index = index
        self.shape = np.shape(data)
        self.nD = len(self.shape)  # 1D,2D,...5D

    def write(self,primary,csysid=''):
        self.write_header(primary,csysid=csysid)
        self.write_index()
        self.write_data()
        #self.f.write('\n')
        
    def write_array(self):
        self.write_header([])
        self.write_data()
        #self.f.write('\n')
        
    def write_text(self, text):
        self.f.write(text+'\n')
        #self.f.write('\n')
        
    def close(self):
        self.f.close()
        
    def boundary(self, BC, arguments, surface):
        table = ''
        for arg in arguments:
            table=table+',%'+arg+'%'
        table = table[1:]
        if BC is 'HGEN':
            self.f.write('bfe,'+surface+','+BC+',1,'+table+'\n\n')    
        elif BC is 'TCC':
            self.f.write('RMODIF,'+surface+',14,'+table+'\n\n')
        else:
            if BC in ['TEMP']:
                BCtype = 'd'
            elif BC in ['CONV','HFLUX']:
                BCtype = 'sf'
            self.f.write(BCtype+','+surface+','+BC+','+table+'\n\n')
        
    def link_args(self, independant):
        for i,var in enumerate(independant):
            self.f.write('*DIM,'+var+',TABLE,1,,,TIME\n')
            self.f.write(var+'(1,1) = arg'+str(i+1)+'\n\n')

    def declare(self,name,shape):
        self.name = name
        self.shape = shape
        self.nD = len(self.shape)
        self.write_header([])
        
    def write_header(self,primary,csysid=''):
        var_list = ['','','',csysid]
        var_str = ''
        for i,var in enumerate(primary):  # primary variables
            var_list[i] = var
        for var in var_list:
            var_str = var_str+','+var
            
        if not primary:  # empty primary array (array)
            array_type = 'ARRAY'
        else:
            array_type = 'TABLE'
            
        self.ijk = np.ones(5)  # upto 5D
        for D in range(self.nD):
            self.ijk[D] = int(self.shape[D])
        if self.nD <= 3:
            self.f.write('*DIM,'+self.name+','+array_type\
                    +',{:1.0f}'.format(self.ijk[0])\
                    +',{:1.0f}'.format(self.ijk[1])\
                    +',{:1.0f}'.format(self.ijk[2])\
                    +var_str+'\n') 
        elif self.nD == 4:
            self.f.write('*DIM,'+self.name+','+array_type[:3]+'4'\
                    +',{:1.0f}'.format(self.ijk[0])\
                    +',{:1.0f}'.format(self.ijk[1])\
                    +',{:1.0f}'.format(self.ijk[2])\
                    +',{:1.0f}'.format(self.ijk[3])\
                    +var_str+'\n') 
        elif self.nD == 5:
            self.f.write('*DIM,'+self.name+','+array_type[:3]+'5'\
                    +',{:1.0f}'.format(self.ijk[0])\
                    +',{:1.0f}'.format(self.ijk[1])\
                    +',{:1.0f}'.format(self.ijk[2])\
                    +',{:1.0f}'.format(self.ijk[3])\
                    +',{:1.0f}'.format(self.ijk[4])\
                    +var_str+'\n')
            
        
    def locate(self, i, cont=1, place=[0,0,0,0,0]):
        loc = '('
        for j in range(self.nD):
            if i == j:
                loc = loc+str(cont)
            else:
                loc = loc+str(place[j])
            if j+1 < self.nD:
                loc = loc+','
        loc = loc+')'
        return loc
        
    def format_row(self, snip):
        row_str = ''        
        for num in snip:
            row_str = row_str+',{:3.7e}'.format(num)
        row_str = row_str[1:]
        return row_str
        
    def list_row(self, line):
        nTen = int(np.floor(len(line)/10))
        count,row_str = 0,[]
        for i in range(nTen):
            row_str.append(self.format_row(line[count:count+10]))
            count += 10
        if len(line) > count:
            row_str.append(self.format_row(line[count:]))
        return row_str
        
    def write_line(self, dimension, line, place=[0,0,0,0,0], TAXIS=0):
        row_str = self.list_row(line)
        cont = 1
        for row in row_str:
            loc = self.locate(dimension, cont=cont, place=place)
            if TAXIS == 1:
                self.f.write('*TAXIS,'+self.name+loc+','+str(dimension+1)+','+row+'\n')
            else:
                self.f.write(self.name+loc+'='+row+'\n')
            cont += 10
      
    def write_index(self):
        #self.f.write('! index\n')
        for i in range(len(self.index)):
            self.write_line(i, self.index[i], place=[1,1,1,1,1], TAXIS=1)
                
    def write_data(self):
        #self.f.write('! data\n')
        for m in range(int(self.ijk[4])):
            for l in range(int(self.ijk[3])):
                for k in range(int(self.ijk[2])):
                    for j in range(int(self.ijk[1])):
                        if self.nD == 1:
                            line = self.data
                        elif self.nD == 2:
                            line = self.data[:,j]
                        elif self.nD == 3:
                            line = self.data[:,j,k]
                        elif self.nD == 4: 
                            line = self.data[:,j,k,l]
                        elif self.nD == 5: 
                            line = self.data[:,j,k,l,m]
                        place = np.array([0,j,k,l,m])+1
                        self.write_line(0, line, place=place)



import itertools
import pickle
import datetime
import os
import numpy as np

def log2lin(x,xlim):
    b = np.log10(xlim[1]/xlim[0])/(xlim[1]-xlim[0])
    a = xlim[1]/10**(b*xlim[1])
    x = a*10**(b*x)
    return x
    
def stripfile(file):
    file = file.replace('.png','')  # strip file type
    file = file.replace('.jpg','')
    return file

class sample_plot(object):
    
    def __init__(self, data, x_origin, y_origin, x_ref, y_ref,
                 x_fig, y_fig, ax_eq, ax, fig, path, file,
                 xscale='linear',yscale='linear'):
        self.xscale = xscale
        self.yscale = yscale
        self.cord_flag = 0
        self.count_data = itertools.count(1)
        self.data = data
        self.path = path
        self.file = file
        self.set_folder()
        self.x = []
        self.y = []
        self.xy = []
        self.x_fig = x_fig
        self.y_fig = y_fig
        self.ax_eq = ax_eq
        self.x_origin = x_origin
        self.y_origin = y_origin
        self.x_ref = x_ref
        self.y_ref = y_ref
        self.ax = ax
        self.fig = fig
        self.cycle_click = itertools.cycle([1,2])
        self.cycle_axis = itertools.cycle([1,2,3,4])
        self.count_click = next(self.cycle_click)
        self.count_axis = next(self.cycle_axis)
        self.cid = x_origin.figure.canvas.mpl_connect('button_press_event', self)
        if self.ax_eq == -1:
            self.count_axis = next(self.cycle_axis)
            self.count_axis = next(self.cycle_axis)
        
    def set_folder(self):
        self.file = stripfile(self.file)
        date_str = datetime.date.today().strftime('%Y_%m_%d')  # today
        self.folder = self.path+'imagedata/'
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        self.file = date_str+'_'+self.file+'.pkl'

    def __call__(self, event):
        if event.button == 1:  # select button
            if self.count_click == 1:  # enter axis
                if self.count_axis == 1:
                    self.x_origin.set_data(event.xdata, event.ydata)
                    self.ax.set_title('select x referance (x='+str(self.x_fig[1])+')')
                    self.x_origin.figure.canvas.draw()
                elif self.count_axis == 2:
                    self.x_ref.set_data(event.xdata, event.ydata)
                    if self.ax_eq == 1:
                        self.ax.set_title('select data (select|finish|undo)')
                        self.count_click = next(self.cycle_click)
                    else:
                        self.ax.set_title('select y-origin (y='+str(self.y_fig[0])+')')   
                    self.x_ref.figure.canvas.draw()                    
                elif self.count_axis == 3:
                    self.y_origin.set_data(event.xdata, event.ydata)
                    self.ax.set_title('select y referance (y='+str(self.y_fig[1])+')')
                    self.y_origin.figure.canvas.draw()
                else:
                    self.y_ref.set_data(event.xdata, event.ydata)
                    self.ax.set_title('select data (select|finish|undo)')
                    self.y_ref.figure.canvas.draw()
                    self.count_click = next(self.cycle_click)
                self.count_axis = next(self.cycle_axis)
            else:  # enter data
                self.x.append(event.xdata)
                self.y.append(event.ydata)
                self.data.set_data(self.x, self.y)
                self.y_ref.figure.canvas.draw()    
        if event.button == 2:  # enter button
            if self.count_click == 2:  # exit and save
                if len(self.x) == 0:
                    self.fig.canvas.mpl_disconnect(self.cid)
                else:
                    if self.cord_flag == 0:
                        self.set_cord()
                        self.cord_flag = 1
                    self.store_data()
                    self.x, self.y = [], []  # reset
                    self.data.set_data(self.x,self.y)
                    self.data.figure.canvas.draw()
        if event.button == 3:  # remove data points
            if self.count_click == 2: 
                if len(self.x) > 0:
                    self.x.pop(len(self.x)-1)
                if len(self.y) > 0:    
                    self.y.pop(len(self.y)-1)
                self.data.set_data(self.x, self.y)
                self.data.figure.canvas.draw() 
                
    def set_cord(self):
        if self.ax_eq == 1:  # referance from x-dir
            x_ref = self.x_ref.get_xydata()[0][0]
            self.x_o = self.x_origin.get_xydata()[0]
            self.x_o[1] = -self.x_o[1]
            self.y_o = self.x_o
            self.x_scale = (self.x_fig[1]-self.x_fig[0])/(x_ref-self.x_o[0])
            self.y_scale = self.x_scale
        elif self.ax_eq == -1:  # referance from y-dir
            y_ref = -self.y_ref.get_xydata()[0][1]
            self.y_o = self.y_origin.get_xydata()[0]
            self.y_o[1] = -self.y_o[1]
            self.x_o = self.y_o
            self.y_scale = (self.y_fig[1]-self.y_fig[0])/(y_ref-self.y_o[1])
            self.x_scale = self.y_scale
        else:  # referance from x and y
            x_ref = self.x_ref.get_xydata()[0][0]
            y_ref = -self.y_ref.get_xydata()[0][1]
            self.x_o = self.x_origin.get_xydata()[0]
            self.x_o[1] = -self.x_o[1]
            self.y_o = self.y_origin.get_xydata()[0]
            self.y_o[1] = -self.y_o[1]
            self.x_scale = (self.x_fig[1]-self.x_fig[0])/(x_ref-self.x_o[0])
            self.y_scale = (self.y_fig[1]-self.y_fig[0])/(y_ref-self.y_o[1])
            
    def store_data(self):
        data = self.data.get_xydata()
        x = data[:,0]
        y = -data[:,1]
        x = self.x_scale*(x-self.x_o[0])+self.x_fig[0]
        y = self.y_scale*(y-self.y_o[1])+self.y_fig[0]
        if self.xscale == 'log':
            x = log2lin(x,self.x_fig)
        if self.yscale == 'log':
            y = log2lin(y,self.y_fig)
        points = {'x':x,'y':y}
        limits = {}
        for var in ['x_o','x_scale','x_fig','y_o','y_scale','y_fig']:
            limits[var] = getattr(self,var)
        with open(self.folder+self.file, 'wb') as output:
                pickle.dump(points,output,-1)
                pickle.dump(data,output,-1)
                pickle.dump(limits,output,-1)
 
def data_mine(path,file,**kw):
    from matplotlib import pyplot as plt
    from amigo.png_tools import sample_plot
    import matplotlib.image as mpimg
    label = kw.get('label','')
    if 'scale' in kw:
        scale = kw.get('scale')
        xscale = scale
        yscale = scale
    else:
        if 'xscale' in kw:
            xscale = kw.get('scale')
        else:
            xscale = 'linear'
        if 'yscale' in kw:
            yscale = kw.get('scale')
        else:
            yscale = 'linear'
    x_fig = kw['xlim']
    y_fig = kw['ylim']
    ax_eq = 0
    
    if len(x_fig) == 0:
        ax_eq = -1
        x_fig = y_fig
        
    if len(y_fig) == 0:
        ax_eq = 1
        y_fig = x_fig

    fig = plt.figure(figsize=(14,14))
    ax = fig.add_subplot(111)
    
    origin = 'upper'
    if '.png' not in file and '.jpg' not in file:
        file += '.png'
    image = mpimg.imread(path+file)
    ax.imshow(image, origin=origin)
    if ax_eq == -1:
        ax.set_title('select y-origin (y='+str(y_fig[0])+')')        
    else:
        ax.set_title('select x-origin (x='+str(x_fig[0])+')')
    
    #default markers
    data, = ax.plot([0],[0],'gs-')
    x_origin, = ax.plot([0],[0],'ro') 
    y_origin, = ax.plot([0],[0],'bo') 
    x_ref, = ax.plot([0],[0],'rx') 
    y_ref, = ax.plot([0],[0],'bx')

    sample_plot = sample_plot(data,x_origin,y_origin,x_ref,y_ref,x_fig,y_fig,
                              ax_eq,ax,fig,path,file+'_'+label,
                              xscale=xscale,yscale=yscale)

def data_load(path,file,**kwargs):
    date = kwargs.get('date',datetime.date.today().strftime('%Y_%m_%d'))
    label = kwargs.get('label','')
    file = stripfile(file)
    with open(path+'imagedata/'+date+'_'+file+'_'+label+'.pkl', 'rb') as input:
                points = pickle.load(input)
                #data = pickle.load(input)
                #limits = pickle.load(input)
    return points








import scipy.optimize as op
import numpy as np
import pylab as pl
import seaborn as sns
import matplotlib

class linelabel(object):

    def __init__(self,MaxLines=100,Ndiv=30,value='1.1f',postfix='',ax='',loc='end'):
        self.ylabel = np.zeros((MaxLines,), dtype=[('x','float'),('y','float'),
                               ('text','|S20'),('color','3float'),
                               ('alpha','float'),('value','|S20'),
                               ('postfix','|S20')])
        self.index = 0
        self.Ndiv = Ndiv
        self.value = value
        self.postfix = postfix
        self.ax = ax
        self.loc = loc
        
    def add(self,label,loc='',value='',postfix=''):
        if not self.ax:
            ax = pl.gca()
        else:
            ax = self.ax
        if len(loc) == 0:
            loc = self.loc
        line = ax.get_lines()[-1]
        color = line.get_color()
        alpha = line.get_alpha()
        if not alpha: alpha = 1
        data = line.get_data()
        if np.max(data[1]) == np.min(data[1]): loc = 'end'
        if loc is 'max':
            n = np.argmax(data[1])
        elif loc is 'min':
            n = np.argmin(data[1])
        elif loc is 'start':
            n = 0
        elif loc is 'xlim':
            xlim = ax.get_xlim()
            n = np.argmin(abs(xlim[-1]-data[0]))
        else:
            n = -1
        x,y = data[0][n],data[1][n]   
        self.ylabel[self.index] = (x,y,label,color,alpha,value,postfix)
        self.index += 1
        
    def space(self,yo):
        y = np.copy(yo)
        y[1:] = yo[0]+np.cumsum(yo[1:])
        #if y[-1] > self.ylim[1]:
        #    y = y-(y[-1]-self.ylim[1])
        #if y[0] < self.ylim[0]:
        #    y = y+(self.ylim[0]-y[0])
        return y
        
    def fit(self,yo,args=()):
        y = self.space(yo)
        err = np.sum((y-args)**2)
        return err
    
    def plot(self,xscale='',yscale='',Ralign=False,Roffset=0):
        fs = matplotlib.rcParams['legend.fontsize']
        if not self.ax:
            ax = pl.gca()
        else:
            ax = self.ax
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        self.ylim = ylim

        if not xscale:
            xscale = ax.get_xscale()
        if not yscale:
            yscale = ax.get_yscale()
        self.ylabel = self.ylabel[:self.index]  # trim
        self.ylabel = np.sort(self.ylabel,order='y')
        if xscale is 'log':
            self.ylabel['x'][self.ylabel['x']==0] = xlim[0]
        if yscale is 'log':
            self.ylabel['y'][self.ylabel['y']==0] = ylim[0]
        yref = self.ylabel['y']
        if xscale is 'log': 
            dx = 10*xlim[1]
        else:
            dx = np.diff(xlim)
            
        dy = np.diff(ylim)/self.Ndiv
        if yscale is 'log': 
            b = np.log10(ylim[1]/ylim[0])/(ylim[1]-ylim[0])
            a = ylim[0]/10**(b*ylim[0])
            yref = np.log10(yref/a)/b
            ylim = np.log10(ylim/a)/b
            #dy = np.log10(dy/a)/b
        
        yo = np.zeros(len(yref))
        yo[0] = yref[0]
        yo[1:] = yref[1:]-yref[:-1]  # spacing
        bounds = [(ylim[0]+dy,None)] 
        #bounds = [(None,None)] 
        for i in range(len(yo)-1):
            bounds.append((dy,None))

        yo = op.minimize(self.fit,yo,args=(yref),
                         bounds=bounds,method='L-BFGS-B',
                         tol=1e-6,options={'disp':False}).x
        y = self.space(yo)

        if yscale is 'log':
            y = a*10**(b*y)

        for i in range(self.index):
            if Ralign:
                xmax = np.max(self.ylabel['x'])+0.06*dx+Roffset
            else:
                xmax = self.ylabel['x'][i]+0.06*dx+Roffset
            ax.plot([self.ylabel['x'][i]+0.01*dx,xmax-0.01*dx],
                    [self.ylabel['y'][i],y[i]],
                    '-',color=self.ylabel['color'][i],
                    linewidth=1,alpha=0.5)  # alpha=self.ylabel['alpha'][i]
            label = self.ylabel['text'][i].decode()        
            value = self.ylabel['value'][i].decode()
            postfix = self.ylabel['postfix'][i].decode()
            if self.value or value:
                if not value: value = self.value
                if not postfix: postfix = self.postfix
                label +=' {:{value}}'.format(self.ylabel['y'][i],
                                             value=value)+postfix
            ax.text(xmax,y[i],label,
                    va='center',color=self.ylabel['color'][i],fontsize=fs,
                    alpha=self.ylabel['alpha'][i])
        #ax.set_xticks(xticks)
        #sns.despine(trim=True)
                

                
                
                


from setuptools import setup

setup(name='amigo',
      version='0.1.1',
      description='equlibrium tools',
      url='https://github.com/twistersi/Amigo',
      author='Simon McIntosh',
      author_email='simon.mcintosh@ukaea.uk',
      license='MIT',
      packages=['amigo'],
			zip_safe=False)



from subprocess import Popen, PIPE
from shlex import split
import os.path

def class_dir(name):
    root = name.__path__[0]
    return root

def trim_dir(check_dir):
    nlevel,dir_found = 3,False
    for i in range(nlevel):
        if os.path.isdir(check_dir):
            dir_found = True
            break
        else:
            if '../' in check_dir:
                check_dir = check_dir.replace('../','',1)
    if not dir_found:
        errtxt = check_dir+' not found\n'
        raise ValueError(errtxt)
    return check_dir

def qsub(script,jobname='analysis',t=1,freiahost=1,
         verbose=True,Code='Nova'):
    p = Popen(split('ssh -T freia{:03.0f}.hpc.l'.format(freiahost)),
              stdin=PIPE,stdout=PIPE,stderr=PIPE)
    p.stdin.write('cd ~/Code/Nova/nova\n'.encode()) 
    
             
    if '.py' not in script:
        script += '.py'
    #wd = 
    py3 = '~/Code/anaconda3/bin/python3'
    flags =  '-V -N {} -j y -wd -S {}'.format(jobname,wd,py3)
    if t > 1:
        flags += ' -t 1:{:1.0f}'.format(t) 
    flags += ' '
    qsub = 'qsub '+flags+script+'\n'
       
    #
    p.stdin.write(qsub.encode())  # submit job  
    p.stdin.flush()
    stdout,stderr = p.communicate()
    if verbose:
        print(stdout.decode(),stderr.decode())

class PATH(object):  # file paths

    def __init__(self,jobname,overwrite=True):
        import datetime
        import os

        self.path = {}
        self.jobname = jobname

        self.date_str = datetime.date.today().strftime('%Y_%m_%d')  # today
        
        # get root dir
        wd = os.getcwd()
        os.chdir('../')
        root = os.getcwd()
        os.chdir(wd)
        
        self.data = self.make_dir(root+'/Data')  # data dir
        if overwrite:
            self.folder = self.rep_dir(self.data+'/'+self.date_str+'.'+self.jobname)
        else:
            self.folder = self.make_dir(self.data+'/'+self.date_str+'.'+self.jobname)
        self.logfile = self.folder+'/run_log.txt'
      
    def new(self, task):
        self.task = task
        if task is 0:
            self.job = self.folder
        else:
            self.job = self.make_dir(self.data+'/'+self.date_str+
                                     '.'+self.jobname)
            self.job = self.rep_dir(self.data+'/'+self.date_str+
                                    '.'+self.jobname+'/task.'+str(task))
        self.screenID = self.jobname+'_task-'+str(task)
        
        # copy config files

        
    def make_dir(self,d):  # check / make
        import os
        if not os.path.exists(d):
            os.makedirs(d)
        return d
            
    def rep_dir(self,d):  # check / replace
        import os
        import shutil as sh
        if os.path.exists(d):
            sh.rmtree(d)
        os.makedirs(d)
        return d

    def go(self):
        import os
        self.home = os.getcwd()
        os.chdir(self.job)

    def back(self):
        import os
        os.chdir(self.home)
        
class SET_PATH(object):  # file paths

    def __init__(self,jobname,**kw):
        import os
        import datetime
        self.os = os
        self.file = {}
        self.jobname = jobname

        if 'date' in kw.keys():
            self.date_str = kw['date']
        else:
            self.date_str = datetime.date.today().strftime('%Y_%m_%d')  # today
        
        # get root dir
        self.home = self.os.getcwd()
        os.chdir('../')
        self.root = os.getcwd()
        os.chdir(self.home)
        
        self.data = self.check_dir(self.root+'/Data')  # data
        self.folder = self.check_dir(self.data+'/'+self.date_str+'.'+jobname)  # job

 
    def check_dir(self,d):  # check / replace
        if not self.os.path.exists(d):
            print(d)
            print('dir not found')
        return d
        
    def go(self,task):
        self.job = self.check_dir(self.data+'/'+self.date_str+'.'+self.jobname+
                                 '/task.'+str(task))  # job
        if 'Rfolder' in self.__dict__.keys():
            self.Rjob = self.make_dir(self.Rfolder+'/task.'+str(task))
        self.os.chdir(self.job)
        
    def back(self):
        self.os.chdir(self.home)
        
    def goto(self,there):
        self.home = self.os.getcwd()
        self.os.chdir(there)
        
    def make_dir(self, d):  # check / make
        import os
        if not os.path.exists(d):
            os.makedirs(d)
        return d
        
    def result(self,task=0,postfix=''):
        import shutil as sh
        self.Rdata = self.make_dir(self.root+'/Results')  # data
        self.Rfolder = self.make_dir(self.Rdata+'/'+
                                     self.date_str+'.'+self.jobname+postfix)    
        # initalise folder + copy inputs

        